##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Views for the grok admin UI"""

import grok
import os
import inspect
from urllib import urlencode

from grokui.admin import docgrok
from grokui.admin.docgrok import DocGrok, DocGrokPackage, DocGrokModule
from grokui.admin.docgrok import DocGrokTextFile, DocGrokGrokApplication
from grokui.admin.docgrok import DocGrokClass, DocGrokInterface, getItemLink

from grokui.admin.objectinfo import ZopeObjectInfo
from grokui.admin.utilities import getPathLinksForObject, getPathLinksForClass
from grokui.admin.utilities import getPathLinksForDottedName, getParentURL
from grokui.admin.utilities import getURLWithParams

from ZODB.broken import Broken
from ZODB.interfaces import IDatabase
from BTrees.OOBTree import OOBTree

import zope.component
from zope.interface import Interface
from zope.interface.interface import InterfaceClass
from zope.app.applicationcontrol.interfaces import IServerControl
from zope.app.applicationcontrol.applicationcontrol import applicationController
from zope.app.applicationcontrol.runtimeinfo import RuntimeInfo
from zope.app.applicationcontrol.browser.runtimeinfo import RuntimeInfoView
from zope.app.applicationcontrol.browser.zodbcontrol import ZODBControlView
from zope.app.apidoc import utilities, codemodule
from zope.app.apidoc.utilities import getPythonPath, renderText, columnize
from zope.app.apidoc.codemodule.module import Module
from zope.app.apidoc.codemodule.class_ import Class
from zope.app.apidoc.codemodule.function import Function
from zope.app.apidoc.codemodule.text import TextFile
from zope.app.apidoc.codemodule.zcml import ZCMLFile
from zope.app.folder.interfaces import IRootFolder
from zope.app.security.interfaces import ILogout, IAuthentication
from zope.app.security.interfaces import IUnauthenticatedPrincipal
from zope.exceptions import DuplicationError
from zope.proxy import removeAllProxies
from zope.tal.taldefs import attrEscape
from ZODB.FileStorage.FileStorage import FileStorageError

import z3c.flashmessage.interfaces


grok.context(IContainer)

class ManageContent(grok.Permission):
    grok.name('meteor.ManageContent')

class Add(grok.View):
    """Add an application.
    """

    grok.require('meteor.ManageContent')

    def update(self, inspectapp=None, application=None):
      """from grokui.admin, maybe not needed for conversion."""
        if inspectapp is not None:
            self.redirect(self.url("docgrok") + "/%s/index"%(
                    application.replace('.','/'),))
        return

    def render(self, content, name, inspectapp=None):
      """from grokui.admin, conversion needed"""
        if name is None or name == "":
            self.redirect(self.url(self.context))
            return
        if name is None or name == "":
            self.redirect(self.url(self.context))
            return
        app = zope.component.getUtility(meteor.interfaces.IContent,
                                        name=content)
        try:
            self.context[name] = app()
            self.flash(u'Added %s `%s`.' % (application, name))
        except DuplicationError:
            self.flash(
                u'Name `%s` already in use. Please choose another name.' % (
                name,))
        self.redirect(self.url(self.context))


class ManageApps(grok.View):
    """Manage applications (delete, rename).
    """

    grok.require('grok.ManageApplications')

    def delete(self, items):
        """Delete applications in items.
        """
        msg = u''
        for name in items:
            try:
                del self.context[name]
                msg = (u'%sApplication `%s` was successfully '
                       u'deleted.\n' % (msg, name))
            except AttributeError:
                # Object is broken.. Try it the hard way...
                # TODO: Try to repair before deleting.
                obj = self.context[name]
                if not hasattr(self.context, 'data'):
                    msg = (
                        u'%sCould not delete application `%s`: no '
                        u'`data` attribute found.\n' % (msg, name))
                    continue
                if not isinstance(self.context.data, OOBTree):
                    msg = (
                        u'%sCould not delete application `%s`: no '
                        u'`data` is not a BTree.\n' % (msg, name))
                    continue
                self.context.data.pop(name)
                self.context.data._p_changed = True
                msg = (u'%sBroken application `%s` was successfully '
                       u'deleted.\n' % (msg, name))

        self.flash(msg)
        self.redirect(self.url(self.context))

    def render(self, rename=None, delete=None, items=None):

        if items is None:
            return self.redirect(self.url(self.context))

        if not isinstance(items, list):
            items = [items]

        if delete is not None:
            return self.delete(items)
        elif rename is not None:
            return self.redirect(getURLWithParams(
                    self.url(self.context, '@@grokadmin_rename'),
                    data=dict(items=items)))
        self.redirect(self.url(self.context))


class MeteorView(grok.View):
    """copied from grokui.admin, should be adapted..

    all below is from previous:

    A grok.View with a special application_url.

    We have to compute the application_url different from common
    grok.Views, because we have no root application object in the
    adminUI. To avoid mismatch, we also call it 'root_url'.

    """

    def root_url(self, name=None):
        obj = self.context
        result = ""
        while obj is not None:
            if IContainer.providedBy(obj):
                return self.url(obj, name)
            obj = obj.__parent__
        raise ValueError("No container nor root element found.")

    def in_docgrok(self):
      """came from grokui.admin, maybe can be removed?"""
        return '/docgrok/' in self.url() or 'inspect.html' in self.url()

    def is_authenticated(self):
        """Check, wether we are authenticated.
        """
        return not IUnauthenticatedPrincipal.providedBy(self.request.principal)


class MeteorMacros(MeteorView):
    """came from grokui.admin, maybe can be removed? wtf is o-wrap?"""
    """Provides the o-wrap layout."""

    grok.context(Interface)


class Rename(MeteorView):
    """Rename Meteor content.
    """
    grok.name('meteor_rename')
    grok.template('rename')
    grok.require('meteor.ManageContent')

    def update(self, cancel=None, items=None, new_names=None):
        msg = u''

        if cancel is not None:
            return self.redirect(self.url(self.context))

        if not isinstance(items, list):
            items = [items]
        self.apps = items

        if new_names is not None and len(new_names) != len(items):
            return self.redirect(self.url(self.context))

        if new_names is None:
            return

        mapping = dict([(items[x], new_names[x]) for x in range(len(items))])

        for oldname, newname in mapping.items():
            if oldname == newname:
                continue
            if oldname not in self.context.keys():
                self.flash('Could not rename %s: not found' % oldname)
                continue
            if newname in self.context.keys():
                self.flash('`%s` already exists.' % newname)
                continue
            self.context[newname] = self.context[oldname]
            self.context[newname].__name__ = newname
            del self.context[oldname]
            self.flash('Renamed `%s` to `%s`.' % (oldname, newname))
        self.redirect(self.url(self.context))
        return

class Index(MeteorView):
    """A redirector to the real frontpage."""

    grok.name('index.html') # The root folder is not a grok.Model
    grok.require('meteor.ManageContent')

    def update(self):
        types = zope.component.getAllUtilitiesRegisteredFor(
            meteor.interfaces.IMeteorContent)
        self.types = ("%s.%s" % (x.__module__, x.__name__)
                             for x in types)
        # Go to the first page immediately.
        self.redirect(self.url('types'))


class Types(MeteorView):
    """View for application management.

    """

    grok.name('typess')
    grok.require('grok.ManageContent')

    def getDocOfApp(self, apppath, headonly = True):
        doctor = docgrok.docgrok_handle(apppath)
        result = doctor.getDoc(headonly)
        if result is None:
            result = ""
        return result

    def update(self):
        from ZODB import broken

        from zope.app.broken.broken import IBroken

        # Available apps...
        apps = zope.component.getAllUtilitiesRegisteredFor(
            grok.interfaces.IApplication)
        self.applications = (
            {'name': "%s.%s" % (x.__module__, x.__name__),
             'docurl':("%s.%s" % (x.__module__, x.__name__)).replace('.', '/')}
            for x in apps)

        # Installed apps...
        inst_apps = [x for x in self.context.values()
                     if hasattr(x, '__class__') and x.__class__ in apps
                     and not issubclass(x.__class__, Broken)]
        inst_apps.sort(lambda x, y: cmp(x.__name__, y.__name__))
        self.installed_applications = inst_apps

        # Broken apps...
        broken_apps = [{'obj':y, 'name':x} for x,y in self.context.items()
                       if isinstance(y, Broken)]
        broken_apps.sort(lambda x, y: cmp(x['name'], y['name']))
        self.broken_applications = broken_apps


class AdminMessageSource(grok.GlobalUtility):

    grok.name('admin')
    zope.interface.implements(z3c.flashmessage.interfaces.IMessageSource)

    message = None

    def send(self, message, type='admin'):
        self.message = z3c.flashmessage.message.PersistentMessage(message,
                                                                  type)

    def list(self, type=None):
        if self.message is None:
            return
        if type is None or self.message.type == type:
            yield self.message

    def delete(self, message):
        if message is self.message:
            self.message = None
        else:
            raise KeyError(message)


class Server(MeteorView, ZODBControlView):
    """Zope3 management screen."""

    grok.require('grok.ManageApplications')

    @property
    def server_control(self):
        return zope.component.queryUtility(IServerControl, '', None)

    @property
    def runtime_info(self):
        riv = RuntimeInfoView()
        riv.context = applicationController
        return riv.runtimeInfo()

    @property
    def current_message(self):
        source = zope.component.getUtility(
          z3c.flashmessage.interfaces.IMessageSource, name='admin')
        messages = list(source.list())
        if messages:
            return messages[0]

    def update(self, time=None, restart=None, shutdown=None,
               admin_message=None, submitted=False,
               dbName="", pack=None, days=0):

        # Packing control
        if pack is not None:
            return self.pack(dbName, days)

        if not submitted:
            return

        # Admin message control
        source = zope.component.getUtility(
          z3c.flashmessage.interfaces.IMessageSource, name='admin')
        if admin_message is not None:
            source.send(admin_message)
        elif getattr(source, 'current_message', False):
            source.delete(source.current_message)

        # Restart control
        if time is not None:
            try:
                time = int(time)
            except:
                time = 0
        else:
            time = 0

        if restart is not None:
            self.server_control.restart(time)
        elif shutdown is not None:
            self.server_control.shutdown(time)

        self.redirect(self.url())

    def pack(self, dbName, days):
        try:
            days = int(days)
        except ValueError:
            self.flash('Error: Invalid Number')
            return
        db = zope.component.getUtility(IDatabase, name=dbName)
        try:
            db.pack(days=days)
            self.flash('ZODB `%s` successfully packed.' % (dbName))
        except FileStorageError, err:
            self.flash('ERROR packing ZODB `%s`: %s' % (dbName, err))


class Users(MeteorView):
    """Users management screen."""

    grok.name('users')
    grok.require('meteor.ManageSite')

    def getPrincipals(self):
        from grok.admin import AUTH_FOLDERNAME, USERFOLDER_NAME

        sm = self.context.getSiteManager()
        if AUTH_FOLDERNAME not in list(sm.keys()):
            return []
        pau = sm[AUTH_FOLDERNAME]
        if USERFOLDER_NAME not in list(pau.keys()):
            return []
        userfolder = pau[USERFOLDER_NAME]
        users = list(userfolder.search({'search':''}))
        return [userfolder.principalInfo(x) for x in users]


    def update(self):
        self.principals = self.getPrincipals()
        pass



def getDottedPathDict(dotted_path):
    """Get a dict containing parts of a dotted path as links.
    """
    if dotted_path is None:
        return {}

    result = []
    part_path = ""
    for part in dotted_path.split('.'):
        name = part
        if part_path != "":
            name = "." + part
        part_path += part
        result.append({
            'name':name,
            'url':"/docgrok/%s" % (part_path,)
            })
        part_path += "/"
    return result

